package ssw.mj.test;

import java.io.StringReader;

import junit.framework.TestCase;

import ssw.mj.Parser;
import ssw.mj.Scanner;
import ssw.mj.Token;

/**
 * Test cases for the <code>Parser</code> class.
 * 
 * @author AW
 */
public class ParserTest extends TestCase {
	private static final String CR = "\r";
	private static final String LF = "\n";
	
	private static final int
		none       = Token.none,
		ident      = Token.ident,
		number     = Token.number,
		charConst  = Token.charConst,
		plus       = Token.plus,
		minus      = Token.minus,
		times      = Token.times,
		slash      = Token.slash,
		rem        = Token.rem,
		eql        = Token.eql,
		neq        = Token.neq,
		gtr        = Token.gtr,
		geq        = Token.geq,
		less       = Token.less,
		leq        = Token.leq,
		and        = Token.and,
		or         = Token.or,
		assign     = Token.assign,
		pplus      = Token.pplus,
		mminus     = Token.mminus,
		semicolon  = Token.semicolon,
		comma      = Token.comma,
		period     = Token.period,
		lpar       = Token.lpar,
		rpar       = Token.rpar,
		lbrack     = Token.lbrack,
		rbrack     = Token.rbrack,
		lbrace     = Token.lbrace,
		rbrace     = Token.rbrace,
		classKW    = Token.classKW,
		continueKW = Token.continueKW,
		doKW       = Token.doKW,
		elseKW     = Token.elseKW,
		finalKW    = Token.finalKW,
		ifKW       = Token.ifKW,
		newKW      = Token.newKW,
		printKW    = Token.printKW,
		readKW     = Token.readKW,
		returnKW   = Token.returnKW,
		voidKW     = Token.voidKW,
		whileKW    = Token.whileKW,
		eof        = Token.eof;

	private static TestPrintWriter output;

	// ---------- helpers ----------
	
	private void initParser(String program) {
		output = new TestPrintWriter();
		Scanner.init(new StringReader(program), output);
	}

	// ----------------------------------------------------
	// ---------- test complete correct programs ----------

	/** Test the shortest valid program: */
	public void testShortestProgram () {
		initParser("class Shortest { }");

		Parser.parse(output);
		output.verify();
	}

	public void testScriptExample () {
		initParser(	"class P" + LF +
					"" + LF +
					"\tfinal int size = 10;" + LF +
					"" + LF +
		/*line  5*/	"\tclass Table {" + LF +
					"\t\tint[] pos;" + LF +
					"\t\tint[] neg;" + LF +
					"\t}" + LF +
					"" + LF +
		/*line 10*/	"\tTable val;" + LF +
					"{" + LF +
					"\tvoid main()" + LF +
					"\t\tint x, i;" + LF +
					"\t{ /*---------- Initialize val */" + LF +
		/*line 15*/	"\t\tval = new Table;" + LF +
					"\t\tval.pos = new int[size]; val.neg = new int[size];" + LF +
					"\t\ti = 0;" + LF +
					"\t\twhile (i < size) {" + LF +
					"\t\t\tval.pos[i] = 0; val.neg[i] = 0;" + LF +
		/*line 20*/	"\t\t\ti++;" + LF +
					"\t\t}" + LF +
					"\t\t/*---------- Read values */" + LF +
					"\t\tread(x);" + LF +
					"\t\twhile (x > 0) {" + LF +
		/*line 25*/	"\t\t\tif (0 <= x && x < size) {" + LF +
					"\t\t\t\tval.pos[x]++;" + LF +
					"\t\t\t} else if (-size < x && x < 0) {" + LF +
					"\t\t\t\tval.neg[-x]++;" + LF +
					"\t\t\t}" + LF +
		/*line 30*/	"\t\t\tread(x);" + LF +
					"\t\t}" + LF +
					"\t}" + LF +
					"}");
		
		Parser.parse(output);
		output.verify();
	}
	
	public void testAllProds () {
		initParser(	"class AllProds" + LF +
					"  final int nPoints = 2;" + LF +
					"  final char exclaim = '!';" + LF +
					"" + LF +
		/*line  5*/	"  class Point {" + LF +
					"    int x, y;" + LF +
					"  }" + LF +
					"" + LF +
					"  char[] msg;" + LF +
		/*line 10*/	"  Point[] pArr;" + LF +
					"" + LF +
					"{" + LF +
					"  void printError ()" + LF +
					"    int i;" + LF +
		/*line 15*/	"  {" + LF +
					"    /* print error message */" + LF +
					"    i = 0;" + LF +
					"    while(i<len(msg)){print(msg[i]);i++;}" + LF +
					"    print(exclaim);" + LF +
		/*line 20*/	"    do {" + LF +
					"      i--;" + LF +
					"      if (i > 5) continue;" + LF +
					"      print(msg[i], 2);" + LF +
					"    } while(i >= 0);" + LF +
		/*line 25*/	"  }" + LF +
					"" + LF +
					"  int getX (Point p) { return p.x; }" + LF +
					"" + LF +
					"  void setX (int x, Point p) {" + LF +
		/*line 30*/	"    if (x > -10 && x <= 9 ||" + LF +
					"        x == 4*3/4%4+12-11*(2147483647/(-2147483647-1)%2) )" + LF +
					"      p.x = x;" + LF +
					"    else printError();" + LF +
					"  }" + LF +
		/*line 35*/	"" + LF +
					"" + LF +
					"  void setMsg () {" + LF +
					"    msg = new char[5];" + LF +
					"    msg[0] = 'E';" + LF +
		/*line 40*/	"    msg[1] = 'r';" + LF +
					"    msg[2] = 'r';" + LF +
					"    msg[3] = 'o';" + LF +
					"    msg[4] = 'r';" + LF +
					"  }" + LF +
		/*line 45*/	"" + LF +
					"  void setPoints ()" + LF +
					"    int i;" + LF +
					"  {" + LF +
					"    pArr = new Point[nPoints];" + LF +
		/*line 50*/	"    i = 0;" + LF +
					"    while (i < len(pArr)) {" + LF +
					"      pArr[i] = new Point;" + LF +
					"      pArr[i].x = i;" + LF +
					"      pArr[i].y = i;" + LF +
		/*line 55*/	"      i++;" + LF +
					"    }" + LF +
					"  }" + LF +
					"" + LF +
					"  void main ()" + LF +
		/*line 60*/	"    int i;" + LF +
					"  {" + LF +
					"    setMsg();" + LF +
					"    setPoints();" + LF +
					"    read(i);" + LF +
		/*line 65*/	"    if (i < nPoints) {" + LF +
					"      print(getX(pArr[i]));" + LF +
					"      print(pArr[i].y);" + LF +
					"    }" + LF +
					"    else" + LF +
		/*line 70*/	"      printError();" + LF +
					"    print(eol);" + LF +
					"  }" + LF +
					"}");

		Parser.parse(output);
		output.verify();
	}
	
	// -----------------------------------------------
	// ---------- test programs with errors ----------

	public void testWrongStart () {
		initParser(	"SomeClass { }");
		
		try { Parser.parse(output); } 
		catch (Error e) {
			if (!e.getMessage().startsWith("PANIC MODE")) throw e;	
		}
		output.addExpectedLine("-- line 1, col 1: class expected");	
		output.verify();
	}
	
	public void testNoProgName () {
		initParser(	"class { }");
		
		try { Parser.parse(output); } 
		catch (Error e) {
			if (!e.getMessage().startsWith("PANIC MODE")) throw e;
		} 
		output.addExpectedLine("-- line 1, col 7: identifier expected");
		output.verify();	
	}

	public void testUndefName () {
		initParser(	"class A" + LF +
					"  integer x;" + LF +
					"{ }");
		
		Parser.parse(output);
		// semantic errors cannot be detected yet
		output.verify();	
	}
	
	// -----------------------------------------------
				
	public static void main (String[] args) {
		junit.textui.TestRunner.run(ParserTest.class);
	}
}
